--EnhancedChatFilter 用户设置
local chatLinesLimit = 20     --多少行内不许重复

--EnhancedChatFilter 内部定义
local ecfVersioin = GetAddOnMetadata("EnhancedChatFilter","Version")
local _, ecf = ...
local utf8replace = ecf.utf8replace -- utf8.lua
local L = ecf.L -- locales.lua
local debugMode = nil
local chatLines = {}
local prevLineID = 0
local filterResult = nil
local filterCharList = "[|@!/<>\"`'_#&;:~\\]"
local filterCharListRegex = "[%(%)%.%%%+%-%*%?%[%]%$%^={}]"

local ecfFrame = CreateFrame("Frame")
local login = nil
local allowWisper = {}

local gsub, select, ipairs, tremove, tinsert, pairs, strsub, getn, sort, format, tonumber, strfind, strlen, strmatch =
gsub, select, ipairs, tremove, tinsert, pairs, strsub, getn, sort, format, tonumber, string.find, strlen, strmatch

--信息中将被过滤器过滤的 UTF-8 符号（常被用于干扰）
local UTF8Symbols = {['·']='',['＠']='',['＃']='',['％']='',
	['＆']='',['＊']='',['——']='',['＋']='',['｜']='',['～']='',['　']='',
	['，']='',['。']='',['、']='',['？']='',['！']='',['：']='',['；']='',
	['’']='',['‘']='',['“']='',['”']='',['【']='',['】']='',['『']='',
	['』']='',['《']='',['》']='',['（']='',['）']='',['￥']='',['＝']='',
	['…']='',['……']='',['１']='1',['２']='2',['３']='3',['４']='4',['５']='5',
	['６']='6',['７']='7',['８']='8',['９']='9',['０']='0',['⒈']='1',['⒉']='2',
	['⒊']='3',['⒋']='4',['⒌']='5',['⒍']='6',['⒎']='7',['⒏']='8',['⒐']='9',
	['〔']='',['〕']='',['〈']='',['〉']='',['‖']=''}

--http://www.wowwiki.com/USERAPI_StringHash
local function StringHash(text)
	local counter = 1
	local len = string.len(text)
	for i = 1, len, 3 do
	counter = math.fmod(counter*8161, 4294967279) +  -- 2^32 - 17: Prime!
		(string.byte(text,i)*16776193) +
		((string.byte(text,i+1) or (len-i+256))*8372226) +
		((string.byte(text,i+2) or (len-i+256))*3932164)
	end
	return math.fmod(counter, 4294967291) -- 2^32 - 5: Prime (and different from the prime in the loop)
end

local EnhancedChatFilter = LibStub("AceAddon-3.0"):NewAddon("EnhancedChatFilter", "AceConsole-3.0", "AceEvent-3.0")
LibStub("AceConfigRegistry-3.0"):NotifyChange("EnhancedChatFilter")

--数据库初始设置
local defaults = {
	profile = {
		--ECF聊天过滤总开关
		enableFilter = true,
		--启用密语白名单过滤
		enableWisper = false,
		--启用“忙碌”玩家过滤
		enableDND = true,
		--启用重复聊天过滤
		enableRPT = true,
		--启用成就刷屏过滤
		enableCFA = true,
		--启用RaidAlert刷屏过滤
		enableRAF = false,
		--启用QuestReport刷屏过滤
		enableQRF = false,
		--启用扩展屏蔽名单
		enableIGM = false,
		--黑名单列表
		blackList = {},
		--扩展屏蔽名单
		ignoreMoreList = {},
		--小地图图标选项
		minimap = {
			hide = false,
			enableMinimapIcon = true,
		}
	}
}

--创建小地图图标数据
local ecfLDB = LibStub("LibDataBroker-1.1"):NewDataObject("Enhanced Chat Filter", {
	type = "data source",
	text = "Enhanced Chat Filter",
	icon = "Interface\\Icons\\Trade_Archaeology_Orc_BloodText",
	OnClick = function() EnhancedChatFilter:EnhancedChatFilterOpen() end,
	OnTooltipShow = function(tooltip)
		tooltip:AddLine("|cffecf0f1Enhanced Chat Filter|r\n"..L["ClickToOpenConfig"]);
	end
})

--创建 Libstub 变量
local icon = LibStub("LibDBIcon-1.0")

--初始化插件
function EnhancedChatFilter:OnInitialize()
	EnhancedChatFilter:RegisterChatCommand("ecf", "EnhancedChatFilterOpen")
	EnhancedChatFilter:RegisterChatCommand("ecf-clear", "EnhancedChatFilterClear")
	EnhancedChatFilter:RegisterChatCommand("ecf-iml", "EnhancedChatFilterIgnoreMoreList")
	EnhancedChatFilter:RegisterChatCommand("ecf-cleariml", "EnhancedChatFilterClearIMList")
	EnhancedChatFilter:RegisterChatCommand("ecf-debug", "EnhancedChatFilterDebug")

	self.db = LibStub("AceDB-3.0"):New("ecfDB", defaults, "Default")
	icon:Register("Enhanced Chat Filter", ecfLDB, self.db.profile.minimap)
end;

--清空关键词确认对话框
StaticPopupDialogs["ECF_CLEAR_BLACKLIST"] = {
	text = L["DoYouWantToClearBlackList?"],
	button1 = L["Yes"],
	button2 = L["No"],
	OnAccept = function() EnhancedChatFilter:EnhancedChatFilterClear() end,
	whileDead = true,
	hideOnEscape = true,
	preferredIndex = 3, -- avoid some UI taint
}

--获取/设置数据库数据的函数
function EnhancedChatFilter:GetEnableFilter()
	return self.db.profile.enableFilter;
end;

function EnhancedChatFilter:ToggleEnableFilter(value)
	self.db.profile.enableFilter = value;
end;

function EnhancedChatFilter:GetEnableWisper()
	return self.db.profile.enableWisper;
end;

function EnhancedChatFilter:ToggleEnableWisper(value)
	self.db.profile.enableWisper = value;
end;

function EnhancedChatFilter:GetEnableDND()
	return self.db.profile.enableDND;
end;

function EnhancedChatFilter:ToggleEnableDND(value)
	self.db.profile.enableDND = value;
end;

function EnhancedChatFilter:GetEnableRPT()
	return self.db.profile.enableRPT;
end;

function EnhancedChatFilter:ToggleEnableRPT(value)
	self.db.profile.enableRPT = value;
end;

function EnhancedChatFilter:GetEnableCFA()
	return self.db.profile.enableCFA;
end;

function EnhancedChatFilter:ToggleEnableCFA(value)
	self.db.profile.enableCFA = value;
end;

function EnhancedChatFilter:GetEnableRAF()
	return self.db.profile.enableRAF;
end;

function EnhancedChatFilter:ToggleEnableRAF(value)
	self.db.profile.enableRAF = value;
end;

function EnhancedChatFilter:GetEnableQRF()
	return self.db.profile.enableQRF;
end;

function EnhancedChatFilter:ToggleEnableQRF(value)
	self.db.profile.enableQRF = value;
end;

function EnhancedChatFilter:GetEnableIGM()
	return self.db.profile.enableIGM;
end;

function EnhancedChatFilter:ToggleEnableIGM(value)
	self.db.profile.enableIGM = value;
end;

function EnhancedChatFilter:GetEnableMinimapIcon()
	return self.db.profile.minimap.enableMinimapIcon;
end;

function EnhancedChatFilter:ToggleEnableMinimapIcon(value)
	self.db.profile.minimap.enableMinimapIcon = value;
end;

function EnhancedChatFilter:GetMinimapStatus()
	return self.db.profile.minimap.hide;
end;

function EnhancedChatFilter:SetMinimapStatus(value)
	self.db.profile.minimap.hide = value;
end;

function EnhancedChatFilter:GetBlackList()
	return self.db.profile.blackList;
end;

function EnhancedChatFilter:InsertBlackList(value, typeModus)
	table.insert(self.db.profile.blackList,{value, typeModus})
end;

function EnhancedChatFilter:RemoveFromBlackList(index)
	table.remove(self.db.profile.blackList,index)
end;

function EnhancedChatFilter:ClearBlackList()
	self.db.profile.blackList = {}
end;

function EnhancedChatFilter:GetIgnoreMoreList()
	return self.db.profile.ignoreMoreList;
end;

function EnhancedChatFilter:InsertIgnoreMoreList(value)
	table.insert(self.db.profile.ignoreMoreList,{value})
end;

function EnhancedChatFilter:RemoveFromIgnoreMoreList(index)
	table.remove(self.db.profile.ignoreMoreList,index)
end;


function EnhancedChatFilter:ClearIgnoreMoreList()
	self.db.profile.ignoreMoreList = {}
end;

--扩展黑名单
local function SendMessage(event, msg)
	local info = ChatTypeInfo[strsub(event, 10)]
	for i = 1, NUM_CHAT_WINDOWS do
		local ChatFrames = _G["ChatFrame"..i]
		if (ChatFrames and ChatFrames:IsEventRegistered(event)) then
			ChatFrames:AddMessage(msg, info.r, info.g, info.b)
		end
	end
end

local function ignoreMore(player)
	if (EnhancedChatFilter:GetEnableIGM() ~= true) then return end
	if (not player) then return end
	local ignore = nil
	if GetNumIgnores() >= 50 then
		for i = 1, GetNumIgnores() do
			local name = GetIgnoreName(i)
			if (player == name) then
				ignore = true
				break
			end
		end
		if (not ignore) then
			local trimmedPlayer = Ambiguate(player, "none")
			EnhancedChatFilter:InsertIgnoreMoreList(trimmedPlayer)
			if debugMode then print("Added to ECF ignoreMoreList!") end
			SendMessage("CHAT_MSG_SYSTEM", format(ERR_IGNORE_ADDED_S, trimmedPlayer))
		end
	end
end

hooksecurefunc("AddIgnore", ignoreMore)
hooksecurefunc("AddOrDelIgnore", ignoreMore)

--禁用暴雪内部语言过滤器
local GetCVar,SetCVar = GetCVar,SetCVar

local frame=CreateFrame("Frame")
frame:Hide()

local function allowSwearing()
	if GetCVar("profanityFilter")~="0" then
		SetCVar("profanityFilter", "0")
		--print("Turned Mature Language Filter off.")
	end
end

frame:SetScript("OnEvent", allowSwearing)
frame:RegisterEvent("VARIABLES_LOADED")
frame:RegisterEvent("CVAR_UPDATE")
frame:RegisterEvent("PLAYER_ENTERING_WORLD")
frame:RegisterEvent("BN_MATURE_LANGUAGE_FILTER")
frame:RegisterEvent("BN_CONNECTED")

allowSwearing() -- 确认再次载入

--登陆/好友列表更新时同步更新允许密语玩家列表
ecfFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
ecfFrame:RegisterEvent("FRIENDLIST_UPDATE")
ecfFrame:SetScript("OnEvent", function(self, event)
	if event == "PLAYER_ENTERING_WORLD" then
		ShowFriends() --强制刷新一次好友列表
	else
		if not login then --登陆时只添加一次
			login = true
			local num = GetNumFriends()
			for i = 1, num do
				local n = GetFriendInfo(i)
				--添加好友到允许名单
				if n then allowWisper[n] = true end
			end
			return
		end
	end
	if debugMode then for k in pairs(allowWisper) do print("allowed: "..k) end end
end)

--自己主动密语的玩家加入安全名单
local function addToAllowWisper(self,_,_,player)
	local trimmedPlayer = Ambiguate(player, "none")
	if allowWisper[trimmedPlayer] then return end --Do nothing if on safe list
	allowWisper[trimmedPlayer] = true --If we want to whisper someone, they're good
end

--拓展黑名单过滤
local function ignoreMoreFilter(self)
	--如果不启用过滤直接返回
	if(EnhancedChatFilter:GetEnableFilter() ~= true) then return end
	if(EnhancedChatFilter:GetEnableIGM() ~= true) then return end

	local ignoreMoreList = EnhancedChatFilter:GetIgnoreMoreList()
	for _,ignorePlayer in ipairs(ignoreMoreList) do
		if (trimmedPlayer == ignorePlayer[1]) then
			if debugMode then print(trimmedPlayer.." Muted~!") end
			return true
		end
	end
end

--密语白名单过滤
local function wisperFilter(self,_,_,player,_,_,_,flags)
	--如果不启用过滤直接返回
	if(EnhancedChatFilter:GetEnableFilter() ~= true) then return end
	if(EnhancedChatFilter:GetEnableWisper() ~= true) then return end

	--避免自己被过滤
	local trimmedPlayer = Ambiguate(player, "none")
	if UnitIsUnit(trimmedPlayer,"player") then return end

	--对 GM 和 DEV 的密语不做处理
	if type(flags) == "string" and (flags == "GM" or flags == "DEV") then return end
	--强制刷新一次好友列表
	ShowFriends()
	--同工会，同团队，同小队和好友不做处理
	if allowWisper[trimmedPlayer] or UnitIsInMyGuild(trimmedPlayer) or UnitInRaid(trimmedPlayer) or UnitInParty(trimmedPlayer) then return end
	--战网好友不做处理
	for i = 1, select(2, BNGetNumFriends()) do
		local GameAccount = BNGetNumFriendGameAccounts(i)
		for j = 1, GameAccount do
			local _, rName, rGame = BNGetFriendGameAccountInfo(i, j)
			if (rName == trimmedPlayer and rGame == "WoW") then return end
		end
	end
	if debugMode then print(player.."'s wisper blocked.") end
	return true
end

--"忙碌"玩家过滤
local function DNDfilter(self,_,_,player,_,_,_,flags)

	--如果不启用过滤直接返回
	if(EnhancedChatFilter:GetEnableFilter() ~= true) then return end
	if(EnhancedChatFilter:GetEnableDND() ~= true) then return end

	--避免自己被过滤
	local trimmedPlayer = Ambiguate(player, "none")
	if UnitIsUnit(trimmedPlayer,"player") then return end

	if type(flags) == "string" and flags == "DND" then return true end
end

--重复信息过滤
local function repeatfilter(self,_,msg,player,_,_,_,_,_,_,_,_,lineID)

	--如果不启用过滤直接返回
	if(EnhancedChatFilter:GetEnableFilter() ~= true) then return end
	if(EnhancedChatFilter:GetEnableRPT() ~= true) then return end

	--如果这行被处理过，则直接引用之前的结果
	if(lineID == prevLineID) then
		return filterResult
	else
		prevLineID = lineID
		filterResult = nil
	end

	--避免自己被过滤
	local trimmedPlayer = Ambiguate(player, "none")
	if UnitIsUnit(trimmedPlayer,"player") then return end

	--删除掉表情和所有标点符号, 但如果对话只有这些内容则保留
	local msgLine = msg:gsub("%p+", ""):gsub("{.*}","")
	if(msgLine == "") then msgLine = msg end

	--处理重复信息
	local msgtable = {Sender = trimmedPlayer, Msg = msgLine, Time = GetTime()}
	for i=1, #chatLines do
		--如果一个人在0.6秒钟之内发了多条信息, 过滤
		--如果一个人的信息完全相同, 过滤
		if (chatLines[i].Sender == msgtable.Sender and ((msgtable.Time - chatLines[i].Time) < 0.600 or chatLines[i].Msg == msgtable.Msg)) then
			chatLines[i] = msgtable
			filterResult = true
			return true
		end
		if i == chatLinesLimit then tremove(chatLines, 1) end
	end
	tinsert(chatLines, msgtable)
end

--关键词过滤
local function blackWordFilter(self,_,msg,player)

	--如果不启用过滤直接返回
	if(EnhancedChatFilter:GetEnableFilter() ~= true) then return end

	--避免自己被过滤
	local trimmedPlayer = Ambiguate(player, "none")
	if UnitIsUnit(trimmedPlayer,"player") then return end

	--从数据库读取黑名单
	local blacklist = EnhancedChatFilter:GetBlackList()
	--if debugMode then print("RAWMsg: "..msg) end

	--删除颜色代码, 物品链接标志, 空格以及一些符号
	local filterString = msg:upper():gsub("|C[0-9A-F]+",""):gsub("|H[^|]+|H",""):gsub("|H|R",""):gsub("%s", ""):gsub(filterCharList, "")

	--删除 UTF-8 干扰符
	filterString = utf8replace(filterString, UTF8Symbols)
	local newfilterString = filterString:gsub(filterCharListRegex, "")

	if debugMode then print("filterString: "..filterString) end

	--从处理过的聊天信息中过滤包含黑名单词语的聊天内容
	for _, blacklistWord in ipairs(blacklist) do
		local keyWord = blacklistWord[1]
		local currentString
		if (blacklistWord[2] == nil) then -- 如果不是正则，也过滤正则中使用的特殊字符
			keyWord = keyWord:upper()
			currentString = newfilterString
		else
			currentString = filterString
		end
		--检查常规黑名单
		if (strfind(currentString,keyWord)) then
			if debugMode then print("Trigger: "..keyWord) end
			return true --deletion of message from chat
		end
	end
end

--团队信息过滤
local function raidWords(self,_,msg,player)
	local RaidAlertTagList = {"%*%*(.+)%*%*", "EUI:(.+)施放了", "EUI:(.+)中断", "EUI:(.+)就绪"}  --RaidAlert 特征
	local QuestReportTagList = {"任务进度提示%s?[:：]", "%(任务完成%)", "<大脚组队提示>", "%[接受任务%]", "<大脚团队提示>"} --QuestReport 特征

	--避免自己被过滤
	local trimmedPlayer = Ambiguate(player, "none")
	if UnitIsUnit(trimmedPlayer,"player") then return end

	if (EnhancedChatFilter:GetEnableFilter() ~= true or not IsInGroup()) then return end
	if (EnhancedChatFilter:GetEnableRAF() == true) then
		for _,RaidAlertTag in ipairs(RaidAlertTagList) do
			if(msg:find(RaidAlertTag)) then
				if debugMode then print("Trigger: "..RaidAlertTag.."in "..msg) end
				return true
			end
		end
	end
	if (EnhancedChatFilter:GetEnableQRF() == true) then
		for _,QuestReportTag in ipairs(QuestReportTagList) do
			if(msg:find(QuestReportTag)) then
				if debugMode then print("Trigger: "..QuestReportTag.."in "..msg) end
				return true
			end
		end
	end
end

ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", addToAllowWisper)

ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER", wisperFilter)

ChatFrame_AddMessageEventFilter("CHAT_MSG_CHANNEL", ignoreMoreFilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_YELL", ignoreMoreFilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_SAY", ignoreMoreFilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER", ignoreMoreFilter)

ChatFrame_AddMessageEventFilter("CHAT_MSG_YELL", DNDfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_SAY", DNDfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER", DNDfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_CHANNEL", DNDfilter)

ChatFrame_AddMessageEventFilter("CHAT_MSG_CHANNEL", blackWordFilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_YELL", blackWordFilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_SAY", blackWordFilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER", blackWordFilter)

ChatFrame_AddMessageEventFilter("CHAT_MSG_SAY", raidWords)
ChatFrame_AddMessageEventFilter("CHAT_MSG_PARTY", raidWords)
ChatFrame_AddMessageEventFilter("CHAT_MSG_PARTY_LEADER", raidWords)
ChatFrame_AddMessageEventFilter("CHAT_MSG_RAID", raidWords)
ChatFrame_AddMessageEventFilter("CHAT_MSG_RAID_LEADER", raidWords)
ChatFrame_AddMessageEventFilter("CHAT_MSG_RAID_WARNING", raidWords)
ChatFrame_AddMessageEventFilter("CHAT_MSG_INSTANCE_CHAT", raidWords)
ChatFrame_AddMessageEventFilter("CHAT_MSG_INSTANCE_CHAT_LEADER", raidWords)

ChatFrame_AddMessageEventFilter("CHAT_MSG_SAY", repeatfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER", repeatfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_CHANNEL", repeatfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_YELL", repeatfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_PARTY", repeatfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_PARTY_LEADER", repeatfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_RAID", repeatfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_RAID_LEADER", repeatfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_RAID_WARNING", repeatfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_INSTANCE_CHAT", repeatfilter)
ChatFrame_AddMessageEventFilter("CHAT_MSG_INSTANCE_CHAT_LEADER", repeatfilter)

---------------------------------------FRAMES---------------------------------------------------------

--main frames
local blackListFrame = CreateFrame("Frame","ConfigFrame",UIParent,"BasicFrameTemplate")
local blackListScroll = CreateFrame("Frame", "ScrollFrame", blackListFrame)

--text
local titleText = blackListFrame:CreateFontString("frametext","ARTWORK","GameFontNormal")
local titleScroll = blackListFrame:CreateFontString("frametext","ARTWORK","GameFontNormal")
local titleAdd = blackListFrame:CreateFontString("frametext","ARTWORK","GameFontNormal")
local toggleText = blackListFrame:CreateFontString("frametext","ARTWORK","GameFontNormal")
local minimapToggleText = blackListFrame:CreateFontString("frametext","ARTWORK","GameFontNormal")
local wisperToggleText = blackListFrame:CreateFontString("frametext","ARTWORK","GameFontNormal")
local dndToggleText = blackListFrame:CreateFontString("frametext","ARTWORK","GameFontNormal")
local rptToggleText = blackListFrame:CreateFontString("frametext","ARTWORK","GameFontNormal")
local cfaToggleText = blackListFrame:CreateFontString("frametext","ARTWORK","GameFontNormal")
local rafToggleText = blackListFrame:CreateFontString("frametext","ARTWORK","GameFontNormal")
local qrfToggleText = blackListFrame:CreateFontString("frametext","ARTWORK","GameFontNormal")
local igmToggleText = blackListFrame:CreateFontString("frametext","ARTWORK","GameFontNormal")
local regexToggleText = blackListFrame:CreateFontString("frametext","ARTWORK","GameFontNormal")

--toggles, input fields, buttons etc
local blackListWordField = CreateFrame("EditBox", "blackListWordField", blackListFrame,"InputBoxTemplate")
local blackListWordIOField = CreateFrame("EditBox", "blackListWordIOField", blackListFrame,"InputBoxTemplate")

local EnableFilterToggle = CreateFrame("CheckButton", "EnableFilterToggle", blackListFrame,"UICheckButtonTemplate")
local EnableMinimap = CreateFrame("CheckButton", "EnableMinimap", blackListFrame,"UICheckButtonTemplate")
local EnableWisper = CreateFrame("CheckButton", "enableWisper", blackListFrame,"UICheckButtonTemplate")
local EnableDND = CreateFrame("CheckButton", "enableDND", blackListFrame,"UICheckButtonTemplate")
local EnableRPT = CreateFrame("CheckButton", "enableRPT", blackListFrame,"UICheckButtonTemplate")
local EnableCFA = CreateFrame("CheckButton", "enableCFA", blackListFrame,"UICheckButtonTemplate")
local EnableRAF = CreateFrame("CheckButton", "enableRAF", blackListFrame,"UICheckButtonTemplate")
local EnableQRF = CreateFrame("CheckButton", "enableQRF", blackListFrame,"UICheckButtonTemplate")
local EnableIGM = CreateFrame("CheckButton", "enableIGM", blackListFrame,"UICheckButtonTemplate")
local EnableRegex = CreateFrame("CheckButton", "enableRegex", blackListFrame,"UICheckButtonTemplate")

local removeButton = CreateFrame("Button", "removeButton", blackListFrame, "UIPanelButtonTemplate")
local addButton = CreateFrame("Button", "addButton", blackListFrame, "UIPanelButtonTemplate")

local arrangeButton = CreateFrame("Button", "arrangeButton", blackListFrame, "UIPanelButtonTemplate")
local clearUpButton = CreateFrame("Button", "clearUpButton", blackListFrame, "UIPanelButtonTemplate")
local importButton = CreateFrame("Button", "importButton", blackListFrame, "UIPanelButtonTemplate")
local exportButton = CreateFrame("Button", "exportButton", blackListFrame, "UIPanelButtonTemplate")

local scrollHighlight = 0
local scrollOffset = 0

local function tablelength(T)
	local count = 0
	for _ in pairs(T) do count = count + 1 end
	return count
end

local function SetToggleText(check, typeToggle)
	local filterType={
		["filter"] = {toggleText, EnableFilterToggle, L["MainFilter"].." |cff2ed689"..L["Enabled"], L["MainFilter"].." |cffE2252D"..L["Disabled"]},
		["minimap"] = {minimapToggleText, EnableMinimap, L["MinimapButton"].." |cff2ed689"..L["Enabled"], L["MinimapButton"].." |cffE2252D"..L["Disabled"]},
		["wisper"] = {wisperToggleText, EnableWisper, L["WhisperWhitelistMode"].." |cff2ed689"..L["Enabled"].."|cffE2252D"..L["BeCareful"], L["WhisperWhitelistMode"].." |cffE2252D"..L["Disabled"]},
		["dnd"] = {dndToggleText, EnableDND, L["DNDfilter"].." |cff2ed689"..L["Enabled"], L["DNDfilter"].." |cffE2252D"..L["Disabled"]},
		["rpt"] = {rptToggleText, EnableRPT, L["RepeatFilter"].." |cff2ed689"..L["Enabled"], L["RepeatFilter"].." |cffE2252D"..L["Disabled"]},
		["cfa"] = {cfaToggleText, EnableCFA, L["AchievementFilter"].." |cff2ed689"..L["Enabled"], L["AchievementFilter"].." |cffE2252D"..L["Disabled"]},
		["raf"] = {rafToggleText, EnableRAF, L["RaidAlertFilter"].." |cff2ed689"..L["Enabled"], L["RaidAlertFilter"].." |cffE2252D"..L["Disabled"]},
		["qrf"] = {qrfToggleText, EnableQRF, L["QuestReportFilter"].." |cff2ed689"..L["Enabled"], L["QuestReportFilter"].." |cffE2252D"..L["Disabled"]},
		["igm"] = {igmToggleText, EnableIGM, L["IgnoreMoreList"].." |cff2ed689"..L["Enabled"], L["IgnoreMoreList"].." |cffE2252D"..L["Disabled"]}
	}
	filterType[typeToggle][1]:SetText(filterType[typeToggle][(check and 3) or 4])
	filterType[typeToggle][2]:SetChecked(check)
end

local function checkBlacklist(blackWord, typeModus) -- 确保关键词自身不会被预处理干扰
	local newWord = blackWord:gsub("%s", ""):gsub(filterCharList, "")
	if (typeModus == nil) then newWord=newWord:gsub(filterCharListRegex, "") end
	newWord = utf8replace(newWord, UTF8Symbols)
	if(newWord ~= blackWord) then return true end
end

local function unlockHighlight()
	for i=1,14 do
		blackListScroll.list[i]:UnlockHighlight()
	end
end

local function BuildUpFrame()
	blackListFrame:SetSize(520,420) -- 主框体
	blackListFrame:SetPoint("Center",UIParent,"CENTER",0,0)
	blackListFrame:SetFrameStrata("HIGH")
	blackListFrame:SetBackdrop({bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background", tile = true, tileSize = 16, edgeSize = 16,insets = { left = 4, right = 4, top = 4, bottom = 4 }})
-- 主框体可以移动 --
	blackListFrame:SetMovable(true)
	blackListFrame:EnableMouse(true)
	blackListFrame:RegisterForDrag("LeftButton")
	blackListFrame:SetScript("OnDragStart", blackListFrame.StartMoving)
	blackListFrame:SetScript("OnDragStop", blackListFrame.StopMovingOrSizing)
-- 鼠标离开插件区域时界面变为半透明 --
	blackListFrame:SetScript("OnUpdate", function(self)
		local current = GetMouseFocus()
		-- Scan through the parent chain
		while current ~= nil do
			if current == self then
				self:SetAlpha(1.0) -- active
				return
			end
			current = current:GetParent()
		end
		self:SetAlpha(0.5) -- inactive
	end)
	blackListFrame:Hide();
-- 主框体可以被Escape关闭 --
	tinsert(UISpecialFrames, blackListFrame:GetName())

	blackListScroll:SetPoint("LEFT",blackListFrame,"LEFT",10,5)
	blackListScroll:SetSize(250,300) -- 关键词列表
	blackListScroll:SetBackdrop({edgeFile = "Interface/Tooltips/UI-Tooltip-Border",tile = true, tileSize = 16, edgeSize = 16,insets = { left = 4, right = 4, top = 4, bottom = 4 }})

	EnableFilterToggle:SetPoint("TOPLEFT",addButton,"BOTTOMLEFT",-6,-6)
	EnableFilterToggle:SetScript("OnClick", function(self)
		EnhancedChatFilter:GetEnableFilter()
		local check = self:GetChecked()
		EnhancedChatFilter:ToggleEnableFilter(check)
		SetToggleText(check, 'filter')
	end)

	EnableMinimap:SetPoint("TOPLEFT",EnableFilterToggle,"BOTTOMLEFT",0,7)
	EnableMinimap:SetScript("OnClick", function(self)
		local check = self:GetChecked()
		EnhancedChatFilter:ToggleEnableMinimapIcon(check)
		SetToggleText(check, 'minimap')
		EnhancedChatFilter:SetMinimapStatus(check)
		if(check) then icon:Show("Enhanced Chat Filter") else icon:Hide("Enhanced Chat Filter") end
	end)

	EnableWisper:SetPoint("TOPLEFT",EnableMinimap,"BOTTOMLEFT",0,7)
	EnableWisper:SetScript("OnClick", function(self)
		local check = self:GetChecked()
		EnhancedChatFilter:ToggleEnableWisper(check)
		SetToggleText(check, 'wisper')
	end)

	EnableDND:SetPoint("TOPLEFT",EnableWisper,"BOTTOMLEFT",0,7)
	EnableDND:SetScript("OnClick", function(self)
		local check = self:GetChecked()
		EnhancedChatFilter:ToggleEnableDND(check)
		SetToggleText(check, 'dnd')
	end)

	EnableRPT:SetPoint("TOPLEFT",EnableDND,"BOTTOMLEFT",0,7)
	EnableRPT:SetScript("OnClick", function(self)
		local check = self:GetChecked()
		EnhancedChatFilter:ToggleEnableRPT(check)
		SetToggleText(check, 'rpt')
	end)

	EnableCFA:SetPoint("TOPLEFT",EnableRPT,"BOTTOMLEFT",0,7)
	EnableCFA:SetScript("OnClick", function(self)
		local check = self:GetChecked()
		EnhancedChatFilter:ToggleEnableCFA(check)
		SetToggleText(check, 'cfa')
	end)

	EnableRAF:SetPoint("TOPLEFT",EnableCFA,"BOTTOMLEFT",0,7)
	EnableRAF:SetScript("OnClick", function(self)
		local check = self:GetChecked()
		EnhancedChatFilter:ToggleEnableRAF(check)
		SetToggleText(check, 'raf')
	end)

	EnableQRF:SetPoint("TOPLEFT",EnableRAF,"BOTTOMLEFT",0,7)
	EnableQRF:SetScript("OnClick", function(self)
		local check = self:GetChecked()
		EnhancedChatFilter:ToggleEnableQRF(check)
		SetToggleText(check, 'qrf')
	end)

	EnableIGM:SetPoint("TOPLEFT",EnableQRF,"BOTTOMLEFT",0,7)
	EnableIGM:SetScript("OnClick", function(self)
		local check = self:GetChecked()
		EnhancedChatFilter:ToggleEnableIGM(check)
		SetToggleText(check, 'igm')
	end)

	EnableRegex:SetPoint("LEFT",blackListWordField,"RIGHT",0,-2)

	blackListWordField:SetPoint("TOPLEFT",ScrollFrame,"TOPRIGHT",30,0) -- 关键词输入区
	blackListWordField:SetSize(180,20)
	blackListWordField:SetMaxLetters(40)
	blackListWordField:SetTextInsets(5, 5, 2, 2)
	blackListWordField:SetText("")
	blackListWordField:SetAutoFocus(false)
	blackListWordField:SetScript("OnTextChanged", function(self,userinput)
		if(userinput == true) then
			if(self:GetText() ~= "") then
				addButton:Enable()
			else
				addButton:Disable()
			end
		end
	end)

	blackListWordIOField:SetPoint("TOPLEFT",removeButton,"BOTTOMLEFT",0,-4) -- 字符串导入导出区
	blackListWordIOField:SetSize(300,20)
	blackListWordIOField:SetTextInsets(5, 5, 2, 2)
	blackListWordIOField:SetText("")
	blackListWordIOField:SetAutoFocus(false)
	blackListWordIOField:SetScript("OnTextChanged", function(self,userinput)
		if(userinput == true) then
			exportButton:Enable()
			local IOnewWord = self:GetText()
			if(IOnewWord ~= "") then
				importButton:Enable()
			else
				importButton:Disable()
			end
		end
	end)

	importButton:SetSize(100,30) -- 导入按钮
	importButton:SetText(L["Import"])
	importButton:SetPoint("LEFT",blackListWordIOField,"RIGHT",2,0)
	importButton:SetScript("OnClick", function()
		local importString = blackListWordIOField:GetText()
		if (strfind(importString, "@")) then
			local oldHashString, newHashString
			importString, newHashString = strsplit("@", importString)
			newHashString = tonumber(newHashString)
			oldHashString = tonumber(StringHash(importString))
			if ( oldHashString ~= newHashString ) then
				print(L["StringHashMismatch"])
				blackListWordIOField:SetFocus()
				return
			else
				print(L["ImportSucceeded"])
			end
		else
			print(L["ImportOldString"])
		end

		local newBlackList = {strsplit(";", importString)}
		for _, blacklist in ipairs(newBlackList) do
			if (blacklist ~= nil) then
				local imNewWord, imTypeWord = strsplit(",",blacklist)
				if (checkBlacklist(imNewWord, imTypeWord) == true) then
					print(imNewWord..L["IncludeAutofilteredWord"])
				else
					EnhancedChatFilter:InsertBlackList(imNewWord,imTypeWord)
				end
			end
		end
		blackListScroll.ScrollFrameUpdate()
		blackListWordIOField:SetText("")
		blackListWordIOField:ClearFocus()
		importButton:Disable()
	end)
	importButton:Disable()

	exportButton:SetSize(100,30) -- 导出按钮
	exportButton:SetText(L["Export"])
	exportButton:SetPoint("LEFT",importButton,"RIGHT",2,0)
	exportButton:SetScript("OnClick", function()
		local blacklist = EnhancedChatFilter:GetBlackList()
		local blackStringList = {}
		for _, blackWord in ipairs(blacklist) do
			if (checkBlacklist(blackWord[1], blackWord[2]) == true) then
				print(blackWord[1]..L["IncludeAutofilteredWord"])
			else
				if (blackWord[2] ~= nil) then
					blackStringList[#blackStringList+1] = blackWord[1]..","..blackWord[2]
				else
					blackStringList[#blackStringList+1] = blackWord[1]
				end
			end
		end
		local blackString = table.concat(blackStringList,";")
		local hashString = StringHash(blackString)
		blackListWordIOField:SetText(blackString.."@"..hashString)
		blackListWordIOField:SetFocus()
		blackListWordIOField:HighlightText()
		exportButton:Disable()
	end)

	addButton:SetSize(205,30) -- 添加关键词按钮
	addButton:SetText(L["AddBlackWordButton"])
	addButton:SetPoint("TOPLEFT",blackListWordField,"BOTTOMLEFT",-2,-4)
	addButton:SetScript("OnClick", function()
		local typeModus = EnableRegex:GetChecked()
		if (typeModus == true) then typeModus = 'regex' else typeModus = nil end
		local newWord = blackListWordField:GetText()
		if (checkBlacklist(newWord, typeModus) == true) then
			print(newWord..L["IncludeAutofilteredWord"])
		else
			EnhancedChatFilter:InsertBlackList(newWord, typeModus)
			blackListScroll.ScrollFrameUpdate()
			blackListWordField:SetText("")
			blackListWordField:ClearFocus()
		end
		addButton:Disable()
		EnableRegex:SetChecked(false)
	end)
	addButton:Disable()

	removeButton:SetSize(80,30) -- 移除关键词按钮
	removeButton:SetText(L["Remove"])
	removeButton:SetPoint("TOPLEFT",ScrollFrame,"BOTTOMLEFT",1,0)
	removeButton:SetScript("OnClick", function()
		unlockHighlight()
		EnhancedChatFilter:RemoveFromBlackList(scrollHighlight)
		scrollHighlight = 0
		blackListScroll.ScrollFrameUpdate()
		removeButton:Disable()
	end)
	removeButton:Disable()

	clearUpButton:SetSize(80,30) -- 清空关键词按钮
	clearUpButton:SetText(L["ClearUp"])
	clearUpButton:SetPoint("TOP",ScrollFrame,"BOTTOM",0,0)
	clearUpButton:SetScript("OnClick", function() StaticPopup_Show("ECF_CLEAR_BLACKLIST") end)

	arrangeButton:SetSize(80,30) -- 整理按钮
	arrangeButton:SetText(L["Arrange"])
	arrangeButton:SetPoint("TOPRIGHT",ScrollFrame,"BOTTOMRIGHT",-1,0)
	arrangeButton:SetScript("OnClick", function()
		local blacklist = EnhancedChatFilter:GetBlackList()
		local hashWord = {}
		local newBlackList = {}

		for _,v in ipairs(blacklist) do
			if (not hashWord[v[1]]) then
				if(v[2] ~= nil and v[2] ~= "regex") then
					print(v[1]..L["SetAsNormalWord"])
					v[2] = nil
				end
				newBlackList[#newBlackList+1] = {v[1], v[2]}
				hashWord[v[1]] = true
			end
		end

		EnhancedChatFilter:ClearBlackList()

		for _,v in ipairs(newBlackList) do
			if (checkBlacklist(v[1], v[2]) == true) then
				print(v[1]..L["ImportOldString"])
			else
				EnhancedChatFilter:InsertBlackList(v[1], v[2])
			end
		end

		blackListScroll.ScrollFrameUpdate()
	end)

	titleText:SetPoint("CENTER", blackListFrame, "TOP", 0, -10)
	titleText:SetJustifyH("left")
	titleText:SetText("Enhanced Chat Filter MOD Fix v"..ecfVersioin)

	toggleText:SetPoint("LEFT",EnableFilterToggle,"RIGHT",5,0)
	toggleText:SetJustifyH("LEFT")
	SetToggleText(EnhancedChatFilter:GetEnableFilter(), 'filter')

	minimapToggleText:SetPoint("LEFT",EnableMinimap,"RIGHT",5,0)
	minimapToggleText:SetJustifyH("LEFT")
	local check = EnhancedChatFilter:GetEnableMinimapIcon()
	SetToggleText(check, 'minimap')
	EnhancedChatFilter:SetMinimapStatus(check)
	if(check) then icon:Show("Enhanced Chat Filter") else icon:Hide("Enhanced Chat Filter") end

	wisperToggleText:SetPoint("LEFT",enableWisper,"RIGHT",5,0)
	wisperToggleText:SetJustifyH("LEFT")
	SetToggleText(EnhancedChatFilter:GetEnableWisper(), 'wisper')

	dndToggleText:SetPoint("LEFT",enableDND,"RIGHT",5,0)
	dndToggleText:SetJustifyH("LEFT")
	SetToggleText(EnhancedChatFilter:GetEnableDND(), 'dnd')

	rptToggleText:SetPoint("LEFT",enableRPT,"RIGHT",5,0)
	rptToggleText:SetJustifyH("LEFT")
	SetToggleText(EnhancedChatFilter:GetEnableRPT(), 'rpt')

	cfaToggleText:SetPoint("LEFT",enableCFA,"RIGHT",5,0)
	cfaToggleText:SetJustifyH("LEFT")
	SetToggleText(EnhancedChatFilter:GetEnableCFA(), 'cfa')

	rafToggleText:SetPoint("LEFT",enableRAF,"RIGHT",5,0)
	rafToggleText:SetJustifyH("LEFT")
	SetToggleText(EnhancedChatFilter:GetEnableRAF(), 'raf')

	qrfToggleText:SetPoint("LEFT",enableQRF,"RIGHT",5,0)
	qrfToggleText:SetJustifyH("LEFT")
	SetToggleText(EnhancedChatFilter:GetEnableQRF(), 'qrf')

	igmToggleText:SetPoint("LEFT",enableIGM,"RIGHT",5,0)
	igmToggleText:SetJustifyH("LEFT")
	SetToggleText(EnhancedChatFilter:GetEnableIGM(), 'igm')

	titleAdd:SetPoint("BOTTOM", blackListWordField, "TOP", 0,6)
	titleAdd:SetJustifyH("center")
	titleAdd:SetText(L["AddBlackWordTitle"])

	titleScroll:SetPoint("BOTTOM", ScrollFrame, "TOP",0,12)
	titleScroll:SetJustifyH("center")
	titleScroll:SetText(L["BlackWordList"])
	
	regexToggleText:SetPoint("BOTTOM", EnableRegex, "TOP",0,0)
	regexToggleText:SetJustifyH("center")
	regexToggleText:SetText(L["Regex"].."?")

	blackListScroll.list = {}
	for i=1,14 do
		blackListScroll.list[i] = CreateFrame("Button",nil,blackListScroll,"SecureHandlerClickTemplate")
		blackListScroll.list[i]:SetSize(200,20)
		blackListScroll.list[i]:SetPoint("TOPLEFT",blackListScroll,"TOPLEFT",8,(i-1)*-20-5)
		blackListScroll.list[i]:SetHighlightTexture("Interface\\Buttons\\UI-PlusButton-Hilight","ADD")
		blackListScroll.list[i]:RegisterForClicks("AnyUp", "AnyDown")

		blackListScroll.list[i].name = blackListScroll.list[i]:CreateFontString(nil,"ARTWORK","GameFontHighlight")
		blackListScroll.list[i].name:SetSize(200-16-8,50)
		blackListScroll.list[i].name:SetJustifyV("CENTER")
		blackListScroll.list[i].name:SetJustifyH("LEFT")

		blackListScroll.list[i]:SetScript("OnClick", function()
			unlockHighlight()
			blackListScroll.list[i]:LockHighlight()
			removeButton:Enable()
			scrollHighlight = i + scrollOffset
		end)
	end
	blackListScroll.scrollFrame = CreateFrame("ScrollFrame","FauxScrollFrameTestScrollFrame",blackListScroll,"FauxScrollFrameTemplate")
	blackListScroll.scrollFrame:SetPoint("TOPLEFT",blackListScroll,"TOPLEFT",0,-6)
	blackListScroll.scrollFrame:SetPoint("BOTTOMRIGHT",blackListScroll,"BOTTOMRIGHT",-28,6)
	blackListScroll.scrollFrame:SetScript("OnShow",blackListScroll.ScrollFrameUpdate)
	blackListScroll.scrollFrame:SetScript("OnVerticalScroll",function(self,offset)
		FauxScrollFrame_OnVerticalScroll(self,offset,20,blackListScroll.ScrollFrameUpdate)
	end)

	local BLZInterfaceAddon = CreateFrame("Frame")

	BLZInterfaceAddon.name = "EnhancedChatFilter"
	InterfaceOptions_AddCategory(BLZInterfaceAddon)
	BLZInterfaceAddon:Hide()

	BLZInterfaceAddon.button = CreateFrame("Button", "BLZInterfaceAddonButton", BLZInterfaceAddon, "UIPanelButtonTemplate")
	BLZInterfaceAddon.button:SetSize(300,25)
	BLZInterfaceAddon.button:SetText("EnhancedChatFilter")
	BLZInterfaceAddon.button:SetPoint("TOP",BLZInterfaceAddon,"TOP",0,-100)
	BLZInterfaceAddon.button:SetScript("OnClick", function ()
		if InterfaceOptionsFrame:IsShown() then
			InterfaceOptionsFrame:Hide()
		end
		blackListFrame:Show()
		GameMenuButtonContinue:Click()
	end)
end

-- create startup event when player logs in
local startUp = CreateFrame("Frame")
startUp:RegisterEvent("PLAYER_LOGIN");

--function that gets called when an event within startUp triggers.
startUp:SetScript("OnEvent", function(self)
	BuildUpFrame()
	print("|cffffff00Enhanced Chat Filter MOD|r|cff3498db v"..ecfVersioin.."|cffecf0f1 - "..L["CommandLine"].." '/ecf'");
end)

function blackListScroll.ScrollFrameUpdate()
	scrollOffset = FauxScrollFrame_GetOffset(blackListScroll.scrollFrame)
	local length=tablelength(EnhancedChatFilter:GetBlackList())
	FauxScrollFrame_Update(blackListScroll.scrollFrame, length, 14, 20)

	unlockHighlight()

	for i=1,14 do
		local idx = scrollOffset+i
		if idx<=length then
			blackListScroll.list[i].name:SetPoint("LEFT",blackListScroll.list[i],"LEFT",2)
			blackListScroll.list[i].name:SetTextColor(1,1,1)
			blackListScroll.list[i].name:SetText(EnhancedChatFilter:GetBlackList()[idx][1])
			if(scrollHighlight == scrollOffset + i) then
				blackListScroll.list[i]:LockHighlight()
			end
			blackListScroll.list[i]:Show()
		else
			blackListScroll.list[i]:Hide()
		end
	end
end

--method run on /ecf
function EnhancedChatFilter:EnhancedChatFilterOpen()
	if(InCombatLockdown()) then return end
	if(blackListFrame:IsVisible() == false) then
		blackListFrame:Show()
	else
		blackListFrame:Hide()
	end
end

--method run on /ecf-clear
function EnhancedChatFilter:EnhancedChatFilterClear()
	EnhancedChatFilter:ClearBlackList()
	blackListScroll.ScrollFrameUpdate()
	print(L["ClearedBlackWordList"])
end

--method run on /ecf-iml
function EnhancedChatFilter:EnhancedChatFilterIgnoreMoreList()
	local ignoreMoreList = EnhancedChatFilter:GetIgnoreMoreList()
	print(L["IgnoreMoreList"]..": ")
	for _,ignorePlayer in ipairs(ignoreMoreList) do
		print(ignorePlayer[1])
	end
end

--method run on /ecf-cleariml
function EnhancedChatFilter:EnhancedChatFilterClearIMList()
	EnhancedChatFilter:ClearIgnoreMoreList()
	print(L["ClearedIgnoreMoreList"])
end

--method run on /ecf-debug
function EnhancedChatFilter:EnhancedChatFilterDebug()
	if(debugMode ~= nil) then
		debugMode=nil
		print("Debug Mode Off!")
	else
		debugMode=true
		print("Debug Mode On!")
	end
end

-----------------------------------------------------------------------
-- ChatFilter 成就刷屏过滤模块
-----------------------------------------------------------------------

local ChatFilter = CreateFrame("Frame")
local achievements, alreadySent = {}, {}
local spamCategories, specialFilters = {[95] = true, [155] = true, [168] = true, [14807] = true, [15165] = true}, {[456] = true, [1400] = true, [1402] = true, [2186] = true, [2187] = true, [2903] = true, [2904] = true, [3004] = true, [3005] = true, [3117] = true, [3259] = true, [3316] = true, [3808] = true, [3809] = true, [3810] = true, [3817] = true, [3818] = true, [3819] = true, [4078] = true, [4079] = true, [4080] = true, [4156] = true, [4576] = true, [4626] = true, [5313] = true, [6954] = true, [7485] = true, [7486] = true, [7487] = true, [8089] = true, [8238] = true, [8246] = true, [8248] = true, [8249] = true, [8260] = true, [8306] = true, [8307] = true, [8398] = true, [8399] = true, [8400] = true, [8401] = true, [8430] = true, [8431] = true, [8432] = true, [8433] = true, [8434] = true, [8435] = true, [8436] = true, [8437] = true, [8438] = true, [8439] = true}

local function SendAchievement(event, achievementID, players)
	if (not players) then return end
	for k in pairs(alreadySent) do alreadySent[k] = nil end
	for i = getn(players), 1, -1 do
		if (alreadySent[players[i].name]) then
			tremove(players, i)
		else
			alreadySent[players[i].name] = true
		end
	end
	if (getn(players) > 1) then
		sort(players, function(a, b) return a.name < b.name end)
	end
	for i = 1, getn(players) do
		local class, color, r, g, b
		if (players[i].guid and players[i].guid:find("Player")) then
			class = select(2, GetPlayerInfoByGUID(players[i].guid))
			color = (CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS)[class]
		end
		if (not color) then
			local info = ChatTypeInfo[strsub(event, 10)]
			r, g, b = info.r, info.g, info.b
		else
			r, g, b = color.r, color.g, color.b
		end
		players[i] = format("|cff%02x%02x%02x|Hplayer:%s|h%s|h|r", r*255, g*255, b*255, players[i].name, players[i].name)
	end
	SendMessage(event, format("[%s]"..L["GotAchievement"].."%s!", table.concat(players, L["And"]), GetAchievementLink(achievementID)))
end

local function achievementReady(id, achievement)
	if (achievement.area and achievement.guild) then
		local playerGuild = GetGuildInfo("player")
		for i = getn(achievement.area), 1, -1 do
			local player = achievement.area[i].name
			if (UnitExists(player) and playerGuild and playerGuild == GetGuildInfo(player)) then
				tinsert(achievement.guild, tremove(achievement.area, i))
			end
		end
	end
	if (achievement.area and getn(achievement.area) > 0) then
		SendAchievement("CHAT_MSG_ACHIEVEMENT", id, achievement.area)
	end
	if (achievement.guild and getn(achievement.guild) > 0) then
		SendAchievement("CHAT_MSG_GUILD_ACHIEVEMENT", id, achievement.guild)
	end
end

local function ChatFrames_OnUpdate(self)
	local found
	for id, achievement in pairs(achievements) do
		if (achievement.timeout <= GetTime()) then
			achievementReady(id, achievement)
			achievements[id] = nil
		end
		found = true
	end
	if (not found) then
		self:SetScript("OnUpdate", nil)
	end
end

local function queueAchievementSpam(event, achievementID, playerdata)
	achievements[achievementID] = achievements[achievementID] or {timeout = GetTime() + 0.5}
	achievements[achievementID][event] = achievements[achievementID][event] or {}
	tinsert(achievements[achievementID][event], playerdata)
	ChatFilter:SetScript("OnUpdate", ChatFrames_OnUpdate)
end

local function ChatFilter_Achievement(self, event, msg, _, _, _, _, _, _, _, _, _, _, guid)
	if EnhancedChatFilter:GetEnableCFA() ~= true then return end
	local achievementID = strmatch(msg, "achievement:(%d+)")
	if (not achievementID) then return end
	achievementID = tonumber(achievementID)
	local Name, Server
	if (guid and guid:find("Player")) then
		Name = select(6, GetPlayerInfoByGUID(guid))
		Server = select(7, GetPlayerInfoByGUID(guid))
		if (Name and Server and strlen(Server) > 0 and Server ~= GetRealmName()) then
			Name = Name.."-"..Server
		end
	end
	local playerdata = {name = Name, guid = guid}
	local categoryID = GetAchievementCategory(achievementID)
	if (spamCategories[categoryID] or spamCategories[select(2, GetCategoryInfo(categoryID))] or specialFilters[achievementID]) then
		queueAchievementSpam((event == "CHAT_MSG_GUILD_ACHIEVEMENT" and "guild" or "area"), achievementID, playerdata)
		return true
	end
end

ChatFrame_AddMessageEventFilter("CHAT_MSG_ACHIEVEMENT", ChatFilter_Achievement)
ChatFrame_AddMessageEventFilter("CHAT_MSG_GUILD_ACHIEVEMENT", ChatFilter_Achievement)

--拾取物品过滤器(阿什兰破碎骨骼+暗月马戏团)
local lootBadWords = {118043, 944, 71096}
local function lootfilter(self,_,msg)
	for _, itemId in ipairs(lootBadWords) do
		if (itemId == tonumber(strmatch(msg, "|Hitem:(%d+)"))) then
			return true
		end
	end
end

ChatFrame_AddMessageEventFilter("CHAT_MSG_LOOT", lootfilter)
